"""
Place whatever executable you like in here and it will be
added to the list of sifters for use in the command.

The expectations of sifters are that the first line output
is the filename to use, and everything else on stdout is
the file to upload to the dashboard.

You can write to stderr without consequence if neccessary,
and returning anything but 0 will cause the upload to be
aborted. Command is run with the following arguments:
<sifter> edx_venv_path edx_platform_path course_id [extra_arg, extra_arg,....]
"""



"""
Tools for use by sifters to assist in setting up
the django environment, output files, etc.
"""

import os
import sys


def use_edx_venv(venv_path):
    """
    For use by a script wanting to use the edx environment.
    It will activate that environment within the script.
    """
    activator = '{0}/bin/activate_this.py'.format(venv_path)
    execfile(activator, dict(__file__=activator))


def enter_lms(venv_path, edx_path):
    """
    This will activate the edx virtual environment, and
    setup the environment as though the script was included
    in the lms project.
    """

    use_edx_venv(venv_path)
    sys.path.append(edx_path)
    os.environ['DJANGO_SETTINGS_MODULE'] = 'lms.envs.aws'
    os.environ['SERVICE_VARIANT'] = 'lms'
    import lms.startup as startup
    startup.run()



"""
xsiftx is a program to run scripts against edx-platform data for a
course or all courses and then writing that to the grade book s3
bucket to show up on the instructor dashboard.

The scripts can be anything that is executable and just need to output
the filename they want to use as the first line, and the file contents
as the rest.  ie:
student_report_20140207.csv
name,datum
johnsmith,things
.
.
.

The script is passed along the course, venv and path to the edx platform
as the first three arguments.  All other arguments passed to xsiftx.py are
passed along as is.
"""

VERSION = "0.1.1"



"""
xsiftx is a program to run scripts against edx-platform data for a
course or all courses and then writing that to the grade book s3
bucket to show up on the instructor dashboard.

The scripts can be anything that is executable and just need to output
the filename they want to use as the first line, and the file contents
as the rest.  ie:
student_report_20140207.csv
name,datum
johnsmith,things
.
.
.

The script is passed the virtualenv path, edx root path, course id, and any
extra arguments that were passed to sifter
"""

import argparse
import os
import StringIO
import subprocess
import sys
import tempfile

import xsiftx.sifters
import xsiftx.store
from xsiftx.util import get_sifters, get_course_list, get_settings


def execute():
    """
    Begin command processing
    """
    sifter_list = get_sifters()
    parser = argparse.ArgumentParser(
        prog='xsiftx.py',
        formatter_class=argparse.RawTextHelpFormatter,
        description=('Run a sifter against one or all courses.\n'
                     'Current available sifters:\n{0}'.format(
                         '\n'.join(sifter_list))
                 )
    )
    parser.add_argument('sifter',
                        help='script in sifter library to run')
    parser.add_argument('-c', '--course', type=str,
                        help='Course ID e.g. org/number/term')
    parser.add_argument('-v', '--venv', type=str, required=True,
                        help='Virtualenv root path for edx-platform')
    parser.add_argument('-e', '--edx-platform', type=str, required=True,
                        help='Root path to edx-platform')

    # Grab any extra arguments passed in
    parser.add_argument('extra_args', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    if args.sifter not in sifter_list:
        sys.stderr.write("You have specified a sifter that doesn't exist\n")
        sys.exit(-1)

    courses = get_course_list(args.venv, args.edx_platform)
    # If course specified, make sure the edx platform has that class
    if args.course:
        if args.course not in courses:
            sys.stderr.write(
                "Course doesn't exist, please pick from:\n{0}\n".format(
                    '\n'.join(courses)
                )
            )
            sys.exit(-1)

    settings = get_settings(args.edx_platform)

    # Everything is all setup, now run the sifter and write the output
    # to the grade download location.
    sifter_path = '{0}/{1}'.format(os.path.dirname(xsiftx.sifters.__file__),
                                   args.sifter)
    data_store = None
    if settings['use_s3']:
        data_store = xsiftx.store.S3Store(settings)
    else:
        data_store = xsiftx.store.FSStore(settings)

    exit_val = 0
    courses_to_run = []
    if args.course:
        courses_to_run = [args.course, ]
    else:
        courses_to_run = courses
    for course in courses_to_run:
        with tempfile.NamedTemporaryFile() as tmpfile:
            with tempfile.NamedTemporaryFile() as stderr_tmp:
                cmd = [sifter_path, args.venv, args.edx_platform, course, ]
                cmd.extend(args.extra_args)
                sift = subprocess.Popen(cmd, stdout=tmpfile, stderr=stderr_tmp,
                                        universal_newlines=True)
                ret_code = sift.wait()
                if ret_code != 0:
                    sys.stderr.write('Sifter {0} called with {1} failed '
                                     'with non zero exit code printing output '
                                     'and aborting\n'.format(args.sifter, cmd))
                    stderr_tmp.flush()
                    stderr_tmp.seek(0)
                    sys.stderr.write(stderr_tmp.read())
                    exit_val = 128
                    continue
                tmpfile.flush()
                tmpfile.seek(0)
                filename = tmpfile.readline()[:-1]
                data_store.store(course, filename, tmpfile)

if __name__ == '__main__':
    execute()



"""
This handles file uploads and hashing for placing the
file on s3 using the edX platform settings
"""
import hashlib
import mimetypes
import os
import urllib

from boto.s3.connection import S3Connection
from boto.s3.key import Key


class FSStore(object):
    """
    This writes out the file to a local path
    """

    def __init__(self, settings):
        self.root_path = settings['root_path']

    def path_for(self, course_id, filename):
        """
        This returns the path to write the file to
        """
        return os.path.join(self.root_path,
                            urllib.quote(course_id, safe=''),
                            filename)

    def store(self, course_id, filename, srcfile):
        """
        Actually writes out the file from wherever srcfile has been
        seeked to.
        """
        full_path = self.path_for(course_id, filename)
        directory = os.path.dirname(full_path)
        if not os.path.exists(directory):
            os.makedirs(directory, 0o755)
        with open(full_path, "wb") as output_file:
            output_file.write(srcfile.read())


class S3Store(object):
    """
    This manages the connection and uploading of files
    generated by the sifter
    """

    def __init__(self, settings):
        self.root_path = settings['root_path']

        conn = S3Connection(
            settings['aws_key_id'],
            settings['aws_key']
        )
        self.bucket = conn.get_bucket(settings['bucket'])

    def key_for(self, course_id, filename):
        """
        Return the S3 key we would use to store and retrive the data for the
        given filename.
        """
        hashed_course_id = hashlib.sha1(course_id)

        key = Key(self.bucket)
        key.key = "{}/{}/{}".format(
            self.root_path,
            hashed_course_id.hexdigest(),
            filename
        )
        return key

    def store(self, course_id, filename, srcfile):
        """
        This actually stores the file into s3
        """

        key = self.key_for(course_id, filename)

        type_guess = mimetypes.guess_type(filename)
        data = srcfile.read()
        key.size = len(data)
        key.content_type = type_guess[0]
        key.content_encoding = type_guess[1]

        key.set_contents_from_string(data)



"""
Package information and dependencies
"""
import os
import sys

from setuptools import setup, find_packages
from distutils.sysconfig import get_python_lib

if "install" in sys.argv:
    lib_paths = [get_python_lib()]
    if lib_paths[0].startswith("/usr/lib/"):
        # We have to try also with an explicit prefix of /usr/local in order to
        # catch Debian's custom user site-packages directory.
        lib_paths.append(get_python_lib(prefix="/usr/local"))
    for lib_path in lib_paths:
        existing_path = os.path.abspath(os.path.join(lib_path, "xsiftx"))

EXCLUDE_FROM_PACKAGES = []

version = __import__('xsiftx').VERSION

setup(
    name='xsiftx',
    version=version,
    url='http://github.com/mitocw/xsiftx',
    author='MITx',
    author_email='mitx-devops@mit.edu',
    description=('Program for running data collection scripts against courses '
                 'and putting the results in the instructor dashboard of the '
                 'local edx-platform instance'),
    license='GNU General Public License v3 (GPLv3)',
    packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    include_package_data=True,
    entry_points={'console_scripts': [
        'xsiftx = xsiftx.command_line:execute',
    ]},
    zip_safe=False,
    install_requires=['boto >= 2.13.3',],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
)



"""
Utility functions for xsiftx.
"""
import json
import os
import stat
import subprocess
import sys

import xsiftx.sifters

ENV_JSON_FILENAME = 'lms.env.json'
AUTH_JSON_FILENAME = 'lms.auth.json'


def get_sifters():
    """
    Get list of currently installed sifters
    """

    sifter_list = []
    sifter_path = os.path.dirname(xsiftx.sifters.__file__)
    executable = stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH
    for filename in os.listdir(sifter_path):
        fullpath = os.path.join(sifter_path, filename)
        if os.path.isfile(fullpath):
            fstat = os.stat(fullpath)
            mode = fstat.st_mode
            if mode & executable:
                sifter_list.append(filename)

    return sifter_list


def get_course_list(venv, edx_root):
    """
    Get a list of courses by using the management commands in edx.
    """

    # Grab course list
    try:
        course_raw = subprocess.check_output(
            ['{0}/bin/python'.format(venv),
             'manage.py', 'lms', '--settings=aws',
             'dump_course_ids', ],
            cwd=edx_root, stderr=subprocess.PIPE
        )
    except OSError as ex:
        sys.stderr.write(
            'No such file or directory: {0!r}\n'.format(str(ex))
        )
        sys.exit(-1)

    except subprocess.CalledProcessError as ex:
        sys.stderr.write(
            'Course listing failed, output was: {0!r}\n'.format(ex.output)
        )
        sys.exit(-1)
    return course_raw.split('\n')[:-1]

def get_settings(edx_root):
    """
    This will pull out the json settings for the
    platform in order to get the bucket, path, key_id, and key
    for uploading to the right place and return them as a dict.
    """
    with open(os.path.abspath('{0}/../{1}'.format(
                edx_root, AUTH_JSON_FILENAME))) as auth_file:
        auth_tokens = json.load(auth_file)

    with open(os.path.abspath('{0}/../{1}'.format(
                edx_root, ENV_JSON_FILENAME))) as env_file:
        env_tokens = json.load(env_file)

    use_s3 = True
    grades_download = env_tokens['GRADES_DOWNLOAD']
    if grades_download['STORAGE_TYPE'] != 'S3':
        use_s3 = False

    bucket = grades_download['BUCKET']
    root_path = grades_download['ROOT_PATH']

    aws_key_id = auth_tokens['AWS_ACCESS_KEY_ID']
    if aws_key_id == '' and use_s3:
        sys.stderr.write('No AWS_ACCESS_KEY_ID\n')
        sys.exit(-2)

    aws_key = auth_tokens["AWS_SECRET_ACCESS_KEY"]
    if aws_key == '' and use_s3:
        sys.stderr.write('No AWS_SECRET_ACCESS_KEY\n')
        sys.exit(-2)

    return dict(aws_key=aws_key, aws_key_id=aws_key_id,
                bucket=bucket, root_path=root_path, use_s3=use_s3)
